import pytest
from freezegun import freeze_time

from cuenca import UserCredential, UserLogin
from cuenca.http import Session


@pytest.mark.vcr
def test_update_password():
    UserCredential.create('222222')
    UserLogin.create('222222')
    UserCredential.update(password='111111')


@pytest.mark.vcr
@freeze_time("2020-03-19")
def test_block_user():
    session = Session()
    session.configure(
        'api_key',
        'api_secret',
        sandbox=True,
        use_jwt=True,
    )
    user_credential = UserCredential.update(
        'US46cuHpEJ5xFTOceMKVqSzF', is_active=False, session=session
    )
    assert not user_credential.is_active
    user_credential = UserCredential.update(
        'US46cuHpEJ5xFTOceMKVqSzF', is_active=True, session=session
    )
    assert user_credential.is_active



import datetime as dt
from typing import ClassVar, List, Optional, cast

from cuenca_validations.types import (
    TransferNetwork,
    TransferQuery,
    TransferRequest,
)
from cuenca_validations.typing import DictStrAny
from pydantic.dataclasses import dataclass
from requests import HTTPError

from ..exc import CuencaException
from .accounts import Account
from .base import Creatable, Transaction
from .resources import retrieve_uri


@dataclass
class Transfer(Transaction, Creatable):
    _resource: ClassVar = 'transfers'
    _query_params: ClassVar = TransferQuery

    updated_at: dt.datetime
    recipient_name: str
    account_number: str
    idempotency_key: str
    network: TransferNetwork
    destination_uri: str
    tracking_key: Optional[str]  # clave rastreo if network is SPEI

    @property  # type: ignore
    def destination(self) -> Optional[Account]:
        return cast(Account, retrieve_uri(self.destination_uri))

    @classmethod
    def create(
        cls,
        account_number: str,
        amount: int,
        descriptor: str,
        recipient_name: str,
        idempotency_key: Optional[str] = None,
    ) -> 'Transfer':
        """
        :param account_number: CLABE
        :param amount: needs to be in centavos (not pesos)
        :param descriptor: how it'll appear for the recipient
        :param recipient_name: name of recipient
        :param idempotency_key: must be unique for each transfer to avoid
            duplicates
        :return: Transfer object

        The recommended idempotency_key scheme:
        1. create a transfer entry in your own database with the status
            created
        2. call this method with the unique id from your database as the
            idempotency_key
        3. update your database with the status created or submitted after
            receiving a response from this method
        """
        if not idempotency_key:
            idempotency_key = cls._gen_idempotency_key(account_number, amount)
        req = TransferRequest(
            account_number=account_number,
            amount=amount,
            descriptor=descriptor,
            recipient_name=recipient_name,
            idempotency_key=idempotency_key,
        )
        return cast('Transfer', cls._create(**req.dict()))

    @classmethod
    def create_many(cls, requests: List[TransferRequest]) -> DictStrAny:
        transfers: DictStrAny = dict(submitted=[], errors=[])
        for req in requests:
            try:
                transfer = cls._create(**req.dict())
            except (CuencaException, HTTPError) as e:
                transfers['errors'].append(dict(request=req, error=e))
            else:
                transfers['submitted'].append(cast('Transfer', transfer))
        return transfers

    @staticmethod
    def _gen_idempotency_key(account_number: str, amount: int) -> str:
        """
        We *strongly* recommend using your own internal database id as the
        idempotency_key, but this provides some level of protection against
        submitting duplicate transfers
        """
        return f'{dt.datetime.utcnow().date()}:{account_number}:{amount}'



import base64
import binascii
import datetime as dt
import json
from dataclasses import dataclass
from typing import TYPE_CHECKING

from .exc import MalformedJwtToken

if TYPE_CHECKING:
    from .http import Session


@dataclass
class Jwt:
    expires_at: dt.datetime
    token: str

    @property
    def is_expired(self) -> bool:
        return self.expires_at - dt.datetime.utcnow() <= dt.timedelta(
            minutes=5
        )

    @staticmethod
    def get_expiration_date(token: str) -> dt.datetime:
        """
        Jwt tokens contains the exp field in the payload data,
        this function extracts the date so we can validate the
        token before any request
        More info about JWT tokens at: https://jwt.io/
        """
        try:
            payload_encoded = token.split('.')[1]
            payload = json.loads(base64.b64decode(f'{payload_encoded}=='))
        except (IndexError, json.JSONDecodeError, binascii.Error):
            raise MalformedJwtToken(f'Invalid JWT: {token}')
        # Expiration timestamp can be found in the `exp` key in the payload
        exp_timestamp = payload['exp']
        return dt.datetime.utcfromtimestamp(exp_timestamp)

    @classmethod
    def create(cls, session: 'Session') -> 'Jwt':
        session.jwt_token = None
        token = session.post('/token', dict())['token']
        expires_at = Jwt.get_expiration_date(token)
        return cls(expires_at, token)



import pytest

from cuenca.resources import BillPayment


@pytest.mark.vcr
def test_bill_payment():
    id_bill_payment = 'ST01'
    bill_payment = BillPayment.retrieve(id_bill_payment)
    assert bill_payment.id == id_bill_payment
    # It should always have a provider
    provider = bill_payment.provider
    assert provider is not None



import pytest
from cuenca_validations.types import ServiceProviderCategory

from cuenca.resources import ServiceProvider


@pytest.mark.vcr
def test_service_provider():
    id_service_provider = 'SP01'
    service_provider = ServiceProvider.retrieve(id_service_provider)
    assert service_provider.id == id_service_provider

    # Check that all categories are in the Category Enum
    categories = service_provider.categories
    values = set(category.value for category in ServiceProviderCategory)
    assert all(category in values for category in categories)



__all__ = [
    'ApiKey',
    'Account',
    'BalanceEntry',
    'BillPayment',
    'Card',
    'CardActivation',
    'CardTransaction',
    'Commission',
    'Deposit',
    'LoginToken',
    'ServiceProvider',
    'Statement',
    'Transfer',
    'UserLogin',
    'WhatsappTransfer',
]

from .accounts import Account
from .api_keys import ApiKey
from .balance_entries import BalanceEntry
from .bill_payments import BillPayment
from .card_activations import CardActivation
from .card_transactions import CardTransaction
from .cards import Card
from .commissions import Commission
from .deposits import Deposit
from .login_tokens import LoginToken
from .resources import RESOURCES
from .service_providers import ServiceProvider
from .statements import Statement
from .transfers import Transfer
from .user_credentials import UserCredential
from .user_logins import UserLogin
from .whatsapp_transfers import WhatsappTransfer

# avoid circular imports
resource_classes = [
    ApiKey,
    Account,
    BalanceEntry,
    BillPayment,
    Card,
    CardActivation,
    CardTransaction,
    Commission,
    Deposit,
    LoginToken,
    ServiceProvider,
    Statement,
    Transfer,
    UserCredential,
    UserLogin,
    WhatsappTransfer,
]
for resource_cls in resource_classes:
    RESOURCES[resource_cls._resource] = resource_cls  # type: ignore



import datetime as dt
from typing import ClassVar, Optional, cast

from cuenca_validations.types import ApiKeyQuery, ApiKeyUpdateRequest
from pydantic.dataclasses import dataclass

from ..http import Session, session as global_session
from .base import Creatable, Queryable, Retrievable, Updateable


@dataclass
class ApiKey(Creatable, Queryable, Retrievable, Updateable):
    _resource: ClassVar = 'api_keys'
    _query_params: ClassVar = ApiKeyQuery

    secret: str
    deactivated_at: Optional[dt.datetime]
    user_id: Optional[str]

    @property
    def active(self) -> bool:
        return (
            self.deactivated_at is None
            or self.deactivated_at > dt.datetime.utcnow()
        )

    @classmethod
    def create(cls, *, session: Session = global_session) -> 'ApiKey':
        return cast('ApiKey', cls._create(session=session))

    @classmethod
    def deactivate(
        cls,
        api_key_id: str,
        minutes: int = 0,
        *,
        session: Session = global_session,
    ) -> 'ApiKey':
        """
        deactivate an ApiKey in a certain number of minutes. If minutes is
        negative, the API will treat it the same as 0. You can't deactivate
        the same key with which the client is configured, since that'd risk
        locking you out. The deactivated key is returned so that you have the
        exact deactivated_at time.
        """
        url = cls._resource + f'/{api_key_id}'
        resp = session.delete(url, dict(minutes=minutes))
        return cast('ApiKey', cls._from_dict(resp))

    @classmethod
    def update(
        cls,
        api_key_id: str,
        metadata: Optional[dict] = None,
        user_id: Optional[str] = None,
        *,
        session: Session = global_session,
    ) -> 'ApiKey':
        """
        If the current user has enough permissions, it associates an ApiKey to
        the `user_id` or updates the correspoding metadata
        """
        req = ApiKeyUpdateRequest(metadata=metadata, user_id=user_id)
        resp = cls._update(api_key_id, **req.dict(), session=session)
        return cast('ApiKey', resp)



from typing import ClassVar, TypeVar, cast

from cuenca_validations.types import EntryType
from pydantic.dataclasses import dataclass

from .accounts import Account
from .base import Queryable, Retrievable, Transaction
from .cards import Card
from .resources import retrieve_uri
from .service_providers import ServiceProvider

FundingInstrument = TypeVar(
    'FundingInstrument', Account, ServiceProvider, Card
)


@dataclass
class BalanceEntry(Retrievable, Queryable):
    _resource: ClassVar = 'balance_entries'

    amount: int  # negative in the case of a debit
    descriptor: str
    name: str
    rolling_balance: int
    type: EntryType
    related_transaction_uri: str
    funding_instrument_uri: str

    @property  # type: ignore
    def related_transaction(self) -> Transaction:
        return cast(Transaction, retrieve_uri(self.related_transaction_uri))

    @property  # type: ignore
    def funding_instrument(self) -> FundingInstrument:
        return cast(
            FundingInstrument,
            retrieve_uri(self.funding_instrument_uri),
        )



__all__ = [
    '__version__',
    'ApiKey',
    'Account',
    'BalanceEntry',
    'BillPayment',
    'Card',
    'CardActivation',
    'CardTransaction',
    'Commission',
    'Deposit',
    'LoginToken',
    'ServiceProvider',
    'Statement',
    'Transfer',
    'UserCredential',
    'UserLogin',
    'WhatsappTransfer',
    'configure',
]


from .http import session
from .resources import (
    Account,
    ApiKey,
    BalanceEntry,
    BillPayment,
    Card,
    CardActivation,
    CardTransaction,
    Commission,
    Deposit,
    LoginToken,
    ServiceProvider,
    Statement,
    Transfer,
    UserCredential,
    UserLogin,
    WhatsappTransfer,
)
from .version import __version__

configure = session.configure



from dataclasses import dataclass

from cuenca_validations.typing import DictStrAny


class CuencaException(Exception):
    ...


class MalformedJwtToken(CuencaException):
    """An invalid JWT token was obtained during authentication"""


class NoResultFound(CuencaException):
    """No results were found"""


class MultipleResultsFound(CuencaException):
    """One result was expected but multiple were returned"""


@dataclass
class CuencaResponseException(CuencaException):
    json: DictStrAny
    status_code: int

    def __str__(self) -> str:
        return repr(self)



from typing import ClassVar

from pydantic.dataclasses import dataclass

from .base import Queryable, Retrievable


@dataclass
class Account(Retrievable, Queryable):
    _resource: ClassVar = 'accounts'

    name: str  # legal name provided by institution
    account_number: str
    institution_name: str



import pytest
from cuenca_validations.types import (
    CardFundingType,
    CardIssuer,
    CardStatus,
    CardType,
)

from cuenca.exc import CuencaResponseException, NoResultFound
from cuenca.resources import Card

user_id = 'US1237'
card_id = 'CAycvo_X9TQoKOKsaAvdqn3w'


@pytest.mark.vcr
def test_card_create():
    card = Card.create(CardIssuer.cuenca, CardFundingType.credit, user_id)
    assert card.id
    assert len(card.number) == 16
    assert card.type == CardType.virtual
    assert card.user_id == user_id


@pytest.mark.vcr
def test_can_not_assign_new_virtual_card():
    with pytest.raises(CuencaResponseException) as exc:
        Card.create(CardIssuer.cuenca, CardFundingType.credit, user_id)
    assert exc.value


@pytest.mark.vcr
def test_card_retrieve():
    card: Card = Card.retrieve(card_id)
    assert card.id == card_id
    assert len(card.number) == 16
    assert card.last_4_digits == '9849'
    assert card.type == CardType.virtual


@pytest.mark.vcr
def test_card_not_found():
    with pytest.raises(CuencaResponseException) as exc:
        Card.retrieve('not-existing-id')
    assert exc.value.status_code == 404
    assert exc.value.json['Code'] == 'NotFoundError'


@pytest.mark.vcr
def test_card_one():
    card = Card.one(
        number='5448750078699849', exp_month=2, exp_year=2026, cvv2='353'
    )
    assert card.id


@pytest.mark.vcr
def test_card_one_errors():
    with pytest.raises(NoResultFound):
        Card.one(user_id='fake id')


@pytest.mark.vcr
def test_card_all():
    cards = Card.all(user_id=user_id)
    assert len(list(cards)) == 1


@pytest.mark.vcr
def test_card_update():
    card = Card.update(card_id, status=CardStatus.blocked)
    assert card.status == CardStatus.blocked
    card = Card.update(card_id, status=CardStatus.active)
    assert card.status == CardStatus.active


@pytest.mark.vcr
def test_deactivate_card():
    card = Card.deactivate(card_id)
    assert card.status == CardStatus.deactivated


@pytest.mark.vcr
def test_card_can_not_be_updated_if_it_is_deactivated():
    with pytest.raises(CuencaResponseException) as exc:
        Card.update(card_id, status=CardStatus.active)
    assert exc.value



import pytest
from cuenca_validations.errors import TooManyAttemptsError
from cuenca_validations.types import CardStatus

from cuenca.resources import CardActivation


@pytest.mark.vcr
def test_card_activation():
    values = dict(
        number='4122943400145596',
        exp_month=11,
        exp_year=24,
        cvv2='123',
    )
    card_activation = CardActivation.create(**values)
    assert card_activation.succeeded
    assert card_activation.user_id == 'US1237'
    card = card_activation.card
    assert all(getattr(card, key) == value for key, value in values.items())
    assert card.user_id == 'US1237'
    assert card.status is CardStatus.active


@pytest.mark.vcr
def test_blocked_attemps():
    values = dict(
        number='4122943499999999',
        exp_month=11,
        exp_year=24,
        cvv2='123',
    )
    for _ in range(6):
        activation = CardActivation.create(**values)
        assert activation.succeeded is False
    with pytest.raises(TooManyAttemptsError):
        CardActivation.create(**values)



import pytest

from cuenca import BalanceEntry


@pytest.mark.vcr
def test_balance_entry_retrieve():
    id_entry = 'TV01'
    balance_entry: BalanceEntry = BalanceEntry.retrieve(id_entry)
    assert balance_entry.id == id_entry
    assert balance_entry.related_transaction
    assert balance_entry.funding_instrument



import json
import os
from typing import Optional, Tuple
from urllib.parse import urljoin

import requests
from cuenca_validations.errors import ERROR_CODES
from cuenca_validations.typing import (
    ClientRequestParams,
    DictStrAny,
    OptionalDict,
)
from requests import Response

from ..exc import CuencaResponseException
from ..jwt import Jwt
from ..version import API_VERSION, CLIENT_VERSION

API_HOST = 'api.cuenca.com'
SANDBOX_HOST = 'sandbox.cuenca.com'


class Session:

    host: str = API_HOST
    basic_auth: Tuple[str, str]
    jwt_token: Optional[Jwt] = None
    session: requests.Session

    def __init__(self):
        self.session = requests.Session()
        self.session.headers.update(
            {
                'X-Cuenca-Api-Version': API_VERSION,
                'User-Agent': f'cuenca-python/{CLIENT_VERSION}',
            }
        )

        # basic auth
        api_key = os.getenv('CUENCA_API_KEY', '')
        api_secret = os.getenv('CUENCA_API_SECRET', '')
        self.basic_auth = (api_key, api_secret)

    @property
    def auth(self) -> Optional[Tuple[str, str]]:
        return self.basic_auth if all(self.basic_auth) else None

    def configure(
        self,
        api_key: Optional[str] = None,
        api_secret: Optional[str] = None,
        use_jwt: Optional[bool] = False,
        sandbox: Optional[bool] = None,
        login_token: Optional[str] = None,
    ):
        """
        This allows us to instantiate the http client when importing the
        client library and configure it later. It's also useful when rolling
        the api key
        """
        if sandbox is False:
            self.host = API_HOST
        elif sandbox is True:
            self.host = SANDBOX_HOST

        # basic auth
        self.basic_auth = (
            api_key or self.basic_auth[0],
            api_secret or self.basic_auth[1],
        )

        if use_jwt:
            self.jwt_token = Jwt.create(self)

        if login_token:
            self.session.headers['X-Cuenca-LoginToken'] = login_token

    def get(
        self, endpoint: str, params: ClientRequestParams = None
    ) -> DictStrAny:
        return self._request_json('get', endpoint, params=params)

    def post(self, endpoint: str, data: DictStrAny) -> DictStrAny:
        return self._request_json('post', endpoint, data=data)

    def patch(self, endpoint: str, data: DictStrAny, **kwargs) -> DictStrAny:
        return self._request_json('patch', endpoint, data=data)

    def delete(self, endpoint: str, data: OptionalDict = None) -> DictStrAny:
        return self._request_json('delete', endpoint, data=data)

    def _request_json(self, *args, **kwargs) -> DictStrAny:
        return json.loads(self.request(*args, **kwargs))

    def request(
        self,
        method: str,
        endpoint: str,
        params: ClientRequestParams = None,
        data: OptionalDict = None,
        **kwargs,
    ) -> bytes:
        if self.jwt_token:
            if self.jwt_token.is_expired:
                self.jwt_token = Jwt.create(self)
            self.session.headers['X-Cuenca-Token'] = self.jwt_token.token
        resp = self.session.request(
            method=method,
            url='https://' + self.host + urljoin('/', endpoint),
            auth=self.auth,
            json=data,
            params=params,
            **kwargs,
        )
        self._check_response(resp)
        return resp.content

    @staticmethod
    def _check_response(response: Response):
        if response.ok:
            return
        json = response.json()
        if 'code' in json:
            raise ERROR_CODES[json['code']](json['error'])
        raise CuencaResponseException(
            json=response.json(),
            status_code=response.status_code,
        )



import pytest

from cuenca import Statement


@pytest.mark.vcr
def test_statement_one_download_pdf():
    statement: Statement = Statement.one(year=2019, month=5)
    assert isinstance(statement.pdf, bytes)


@pytest.mark.vcr
def test_statement_one_download_xml():
    statement: Statement = Statement.one(year=2019, month=7)
    assert isinstance(statement.xml, bytes)
